"""Constants - Application global constants."""

from typing import Final

APP_NAME: Final = "Tony"
VERSION: Final = "1.6.1"
PACKAGE_NAME: Final = "neuro-api-tony"
GIT_REPO_URL: Final = "https://github.com/Pasu4/neuro-api-tony"
PYPI_URL: Final = "https://pypi.org/project/neuro-api-tony"
PYPI_API_URL: Final = "https://pypi.org/pypi/neuro-api-tony/json"
